import os
import sys

# Add package to path.
file_dir = os.path.dirname(__file__)
sys.path.insert(0, os.path.abspath(os.path.join(file_dir, '..')))

# Load LAB extensions.
# noinspection PyUnresolvedReferences
import lab.torch
# noinspection PyUnresolvedReferences
import lab.tensorflow

# Load TensorFlow extension.
# noinspection PyUnresolvedReferences
import varz.tensorflow



import logging

from autograd import value_and_grad

from ..minimise import make_l_bfgs_b, make_adam, exception

__all__ = ['minimise_l_bfgs_b', 'minimise_adam']

log = logging.getLogger(__name__)


def _wrap_f(vs, names, f):
    # Differentiable assignments will overwrite the variables, so make a copy.
    vs_copy = vs.copy()

    # Keep track of function evaluations.
    f_evals = []

    def f_vectorised(x):
        vs_copy.set_vector(x, *names, differentiable=True)
        return f(vs_copy)

    def f_wrapped(x):
        # Compute objective function value.
        try:
            obj_value, grad = value_and_grad(f_vectorised)(x)
        except Exception as e:
            return exception(x, e)

        f_evals.append(obj_value)
        return obj_value, grad

    return f_evals, f_wrapped


minimise_l_bfgs_b = make_l_bfgs_b(_wrap_f)
minimise_adam = make_adam(_wrap_f)



import importlib
import logging
import re
from itertools import product

import lab as B
from plum import Referentiable, Dispatcher, Self

__all__ = ['lazy_tf', 'lazy_torch', 'Initialiser', 'Packer', 'match']

log = logging.getLogger(__name__)


class LazyModule:
    """A module that loads once an attribute is requested.

    Args:
        name (str): Name of module.
    """

    def __init__(self, name):
        self.name = name
        self.module = None

    def __getattr__(self, item):
        # Load module.
        if self.module is None:
            self.module = importlib.import_module(self.name)

        return getattr(self.module, item)


lazy_tf = LazyModule('tensorflow')
lazy_torch = LazyModule('torch')


class Initialiser:
    """Variable initialiser."""

    def __init__(self):
        self._assignments = {}

    def assign(self, name, values):
        """Assign values to a particular variable.

        Args:
            name (str): Name of variables.
            values (list[tensor]): List of values to assign.
        """
        self._assignments[name] = values

    def generate(self, vs):
        """Generate initialisers.

        Args:
            vs (:class:`.vars.Vars`): Variable storage.

        Returns:
            list[function]: List of functions that perform the initialisations.
        """
        names, value_sets = zip(*self._assignments.items())
        return [_construct_assigner(vs, names, values)
                for values in product(*value_sets)]


def _construct_assigner(vs, names, values):
    def assign():
        return [vs.assign(name, val) for name, val in zip(names, values)]

    return assign


class Packer(metaclass=Referentiable):
    """Pack objects into a vector.

    Args:
        *objs (tensor): Objects to pack.
    """
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch([object])
    def __init__(self, *objs):
        self._shapes = [B.shape(obj) for obj in objs]
        self._lengths = [B.length(obj) for obj in objs]

    @_dispatch({tuple, list})
    def __init__(self, objs):
        Packer.__init__(self, *objs)

    @_dispatch([object])
    def pack(self, *objs):
        """Pack objects.

        Args:
            *objs (tensor): Objects to pack.

        Returns:
            tensor: Vector representation of the objects.
        """
        return B.concat(*[B.flatten(obj) for obj in objs], axis=0)

    @_dispatch({tuple, list})
    def pack(self, objs):
        return self.pack(*objs)

    def unpack(self, package):
        """Unpack vector.

        Args:
            package (tensor): Vector to unpack.

        Returns:
            list[tensor]: Original objects.
        """
        i, outs = 0, []
        for shape, length in zip(self._shapes, self._lengths):
            outs.append(B.reshape(package[i:i + length], *shape))
            i += length
        return outs


def match(pattern, target):
    """Match a string exactly against a pattern, where the pattern is a regex
    with '*' as the only active character.

    Args:
        pattern (str): Pattern.
        target (str): Target.

    bool: `True` if `pattern` matches `target`.
    """
    pattern = ''.join('.*' if c == '*' else re.escape(c) for c in pattern)
    return bool(re.match('^' + pattern + '$', target))



import lab as B
import numpy as np
import pytest
import tensorflow as tf
import torch
import varz.autograd
import varz.tensorflow
import varz.torch
from varz import Vars

from .util import approx, Value, OutStream


@pytest.fixture(params=[(np.float64, varz.autograd.minimise_l_bfgs_b, {}),
                        (torch.float64, varz.torch.minimise_l_bfgs_b, {}),
                        (tf.float64, varz.tensorflow.minimise_l_bfgs_b, {}),
                        (np.float64, varz.autograd.minimise_adam,
                         {'rate': 1e-1}),
                        (torch.float64, varz.torch.minimise_adam,
                         {'rate': 1e-1}),
                        (tf.float64, varz.tensorflow.minimise_adam,
                         {'rate': 1e-1})])
def minimise_method(request):
    yield request.param


def test_minimise(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    # Initialise a variable that is not used.
    vs.get(name='other')

    # Define some objective.
    def f(vs_):
        return (-3 - vs_.pos(name='x', init=5.)) ** 2

    # Minimise it, until convergence.
    val_opt = minimise(f, vs, iters=10000, **kw_args)

    # Check for equality up to three digits.
    approx(val_opt, 9, digits=3)
    approx(vs['x'], 0, digits=3)


def test_minimise_disconnected_gradient(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)
    vs.get(name='x')

    # Check that optimiser runs for objective that returns the constant zero.
    minimise(lambda v: B.cast(v.dtype, 0), vs, **kw_args)


def test_minimise_exception(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    first_call = Value(True)

    # Define an objective that sometimes fails after the first call.
    def f(vs_):
        if first_call.val:
            first_call.val = False
        else:
            if np.random.rand() > .5:
                raise Exception('Fail!')
        return vs_.get(name='x', init=5.) ** 2

    # Check that the optimiser runs.
    minimise(f, vs, **kw_args)


def test_minimise_zero_calls(minimise_method):
    dtype, minimise, kw_args = minimise_method
    vs = Vars(dtype=dtype)

    calls = Value(0)

    def f(vs_):
        calls.val += 1
        return vs_.get(name='x', init=5.) ** 2

    # Check that running the optimiser for zero iterations only incurs a
    # single call.
    minimise(f, vs, iters=0, **kw_args)
    assert calls.val == 1


def test_minimise_trace(minimise_method):
    dtype, minimise, kw_args = minimise_method

    def f(vs_):
        return vs_.get(name='x') ** 2

    # Test that `trace=False` prints nothing.
    with OutStream() as stream:
        minimise(f, Vars(dtype=dtype), trace=False, **kw_args)
        assert stream.output == ''

    # Test that `trace=False` prints something.
    with OutStream() as stream:
        minimise(f, Vars(dtype=dtype), trace=True, **kw_args)
        assert stream.output != ''



import logging

import lab as B
import numpy as np
import wbml.out as out
from plum import Dispatcher
from scipy.optimize import fmin_l_bfgs_b

__all__ = ['minimise_l_bfgs_b', 'minimise_adam']

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


def minimise_l_bfgs_b(f,
                      vs,
                      f_calls=10000,
                      iters=1000,
                      trace=False,
                      names=None):  # pragma: no cover
    """Minimise a function with L-BFGS-B.

    Args:
        f (function): Function to optimise.
        vs (:class:`.vars.Vars`): Variable manager.
        f_calls (int, optional): Maximum number of function calls. Defaults to
            `10000`.
        iters (int, optional): Maximum number of iterations. Defaults to
            `1000`.
        trace (bool, optional): Show trace of optimisation. Defaults to `False`.
        names (list, optional): List of names of variables to optimise. Defaults
            to all variables.

    Returns:
        float: Final objective function value.
    """
    raise RuntimeError('Call a backend-specific optimiser instead.')


def minimise_adam(f,
                  vs,
                  iters=1000,
                  rate=1e-3,
                  beta1=0.9,
                  beta2=0.999,
                  epsilon=1e-8,
                  trace=False,
                  names=None):  # pragma: no cover
    """Minimise a function with Adam.

    Args:
        f (function): Function to optimise.
        vs (:class:`.vars.Vars`): Variable manager.
        iters (int, optional): Maximum number of iterations. Defaults to
            `1000`.
        rate (float, optional): Learning rate. Defaults to `1e-3`.
        beta1 (float, optional): Exponential decay for mean. Defaults to `0.9`.
        beta2 (float, optional): Exponential decay for second moment.
            Defaults to `0.999`.
        epsilon (float, optional): Small value to prevent division by zero.
            Defaults to `1e-8`.
        trace (bool, optional): Show trace of optimisation. Defaults to `False`.
        names (list, optional): List of names of variables to optimise. Defaults
            to all variables.

    Returns:
        float: Final objective function value.
    """
    raise RuntimeError('Call a backend-specific optimiser instead.')


def make_l_bfgs_b(wrap_f):
    """Create `minimise_l_bfgs_b` given a function wrapper.

    Args:
        wrap_f (function): Function wrapper.

    Returns:
        function: `minimise_l_bfgs_b`.
    """

    def minimise_l_bfgs_b(f,
                          vs,
                          f_calls=10000,
                          iters=1000,
                          trace=False,
                          names=None):
        names = [] if names is None else names

        # Run function once to ensure that all variables are initialised and
        # available.
        val_init = f(vs)

        # SciPy doesn't perform zero iterations, so handle that edge case
        # manually.
        if iters == 0 or f_calls == 0:
            return B.to_numpy(val_init)

        # Extract initial value.
        x0 = B.to_numpy(vs.get_vector(*names))

        # Wrap the function and get the list of function evaluations.
        f_vals, f_wrapped = wrap_f(vs, names, f)

        # Perform optimisation routine.
        def perform_minimisation(callback_=lambda _: None):
            return fmin_l_bfgs_b(func=f_wrapped,
                                 x0=x0,
                                 maxiter=iters,
                                 maxfun=f_calls,
                                 callback=callback_,
                                 disp=0)

        if trace:
            # Print progress during minimisation.
            with out.Progress(name='Minimisation of "{}"'.format(f.__name__),
                              total=iters) as progress:
                def callback(_):
                    progress({'Objective value': np.min(f_vals)})

                x_opt, val_opt, info = perform_minimisation(callback)

            with out.Section('Termination message'):
                out.out(info['task'].decode('utf-8'))
        else:
            # Don't print progress; simply perform minimisation.
            x_opt, val_opt, info = perform_minimisation()

        vs.set_vector(x_opt, *names)  # Assign optimum.

        return val_opt  # Return optimal value.

    return minimise_l_bfgs_b


def make_adam(wrap_f):
    """Create `minimise_adam` given a function wrapper.

    Args:
        wrap_f (function): Function wrapper.

    Returns:
        function: `minimise_adam`.
    """

    def minimise_adam(f,
                      vs,
                      iters=1000,
                      rate=1e-3,
                      beta1=0.9,
                      beta2=0.999,
                      epsilon=1e-8,
                      trace=False,
                      names=None):
        names = [] if names is None else names

        # Run function once to ensure that all variables are initialised and
        # available.
        val_init = f(vs)

        # Handle the edge case of zero iterations.
        if iters == 0:
            return B.to_numpy(val_init)

        # Extract initial value.
        x0 = B.to_numpy(vs.get_vector(*names))

        # Wrap the function.
        _, f_wrapped = wrap_f(vs, names, f)

        def perform_minimisation(callback_=lambda _: None):
            # Perform optimisation routine.
            x = x0
            obj_value = None
            m = np.zeros_like(x0)
            v = np.zeros_like(x0)

            for i in range(iters):
                obj_value, grad = f_wrapped(x)

                callback_(obj_value)

                # Update estimates of moments.
                m = beta1 * m + (1 - beta1) * grad
                v = beta2 * v + (1 - beta2) * grad ** 2

                # Correct for bias of initialisation.
                m_corr = m / (1 - beta1 ** (i + 1))
                v_corr = v / (1 - beta2 ** (i + 1))

                # Perform update.
                x = x - rate * m_corr / (v_corr ** .5 + epsilon)

            return x, obj_value

        if trace:
            # Print progress during minimisation.
            with out.Progress(name='Minimisation of "{}"'.format(f.__name__),
                              total=iters) as progress:
                def callback(obj_value):
                    progress({'Objective value': obj_value})

                x_opt, obj_value = perform_minimisation(callback)
        else:
            x_opt, obj_value = perform_minimisation()

        vs.set_vector(x_opt, *names)  # Assign optimum.

        return obj_value  # Return last objective value.

    return minimise_adam


def exception(x, e):
    """In the case that an exception is raised during function evaluation,
    print a warning and return NaN for the function value and gradient.

    Args:
        x (tensor): Current input.
        e (:class:`Exception`): Caught exception.

    Returns:
        tuple: Tuple containing NaN and NaNs for the gradient.
    """
    log.warning('Caught exception during function evaluation: '
                '"{}". Returning NaN.'.format(e))
    grad_nan = np.empty(x.shape)
    grad_nan[:] = np.nan
    return np.nan, grad_nan



import inspect
from abc import ABCMeta, abstractmethod
from functools import wraps
from inspect import isfunction

from .vars import Provider

__all__ = ['sequential',
           'Unbounded',
           'Positive',
           'Bounded',
           'LowerTriangular',
           'PositiveDefinite',
           'Orthogonal',
           'parametrised']


class Sequential(Provider):
    """A variable provider that wraps a :class:`.vars.Vars` object and
    automatically names unnamed variables sequentially (0, 1, 2, ...) with a
    possible prefix.

    Args:
        vs (:class:`.vs.Vars`): Variable container to wrap.
        prefix (str): Prefix for names in the sequence.
    """

    def __init__(self, vs, prefix):
        self.vs = vs
        self.prefix = prefix
        self.count = 0

    def _get_var(self, getter, args, kw_args):
        if 'name' not in kw_args:
            kw_args['name'] = f'{self.prefix}{self.count}'
            self.count += 1
        return getter(*args, **kw_args)

    def unbounded(self, *args, **kw_args):
        return self._get_var(self.vs.unbounded, args, kw_args)

    def positive(self, *args, **kw_args):
        return self._get_var(self.vs.positive, args, kw_args)

    def bounded(self, *args, **kw_args):
        return self._get_var(self.vs.bounded, args, kw_args)

    def lower_triangular(self, *args, **kw_args):
        return self._get_var(self.vs.lower_triangular, args, kw_args)

    def positive_definite(self, *args, **kw_args):
        return self._get_var(self.vs.positive_definite, args, kw_args)

    def orthogonal(self, *args, **kw_args):
        return self._get_var(self.vs.orthogonal, args, kw_args)

    def __getitem__(self, name):
        return self.vs[name]


def _to_sequential(x, prefix):
    """Convert a variable provider to sequential, if it is one.

    Args:
        x (object): Object to convert.
        prefix (str): Prefix for names in the sequence.

    Returns:
        object: `x` converted to sequential if it is a variable provider,
            otherwise just `x`.
    """
    return Sequential(x, prefix) if isinstance(x, Provider) else x


def _extract_prefix_and_f(prefix_or_f):
    """Extract the prefix and function.

    Args:
        prefix_or_f (object): Either a prefix or a function.

    Returns:
        tuple: Tuple containing the predict and the function. If a function
            was given, the prefix defaults to an empty string; and if a prefix
            was given, the function defaults to `None`.
    """
    if prefix_or_f is None:
        # Not used as decorator and prefix is left unspecified.
        return '', None
    elif isfunction(prefix_or_f):
        # Used as a decorator.
        return '', prefix_or_f
    else:
        # Not used as decorator and prefix is specified.
        return prefix_or_f, None


def sequential(prefix=None):
    """Decorator that generates variable names for unnamed variables
    sequentially.

    Args:
        prefix (str, optional): Prefix to prepend to the name of generated
            variables. Defaults to no prefix.
    """
    prefix, f = _extract_prefix_and_f(prefix)

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            # Replace all variable containers with their sequential variants.
            args = [_to_sequential(x, prefix) for x in args]
            kw_args = {k: _to_sequential(v, prefix) for k, v in kw_args.items()}

            return f_(*args, **kw_args)

        return wrapped_f

    # Return the decorated function directly if a function was given.
    return decorator if f is None else decorator(f)


class VariableType(metaclass=ABCMeta):
    """A type of a variable. Any arguments are passed to the appropriate method
    of :class:`.vars.Vars` to instantiate the variable."""

    def __init__(self, *args, **kw_args):
        self.args = args
        self.kw_args = kw_args

    def _get_var(self, method, name, init):
        # Copy dict to not modify the original.
        kw_args = dict(self.kw_args)

        # Set initial value.
        if (
                init is not None and
                'init' in kw_args and
                kw_args['init'] is not None
        ):
            raise ValueError(f'Initial value doubly specified: '
                             f'"{init}" and "{kw_args["init"]}".')
        kw_args['init'] = init

        # Set name.
        if 'name' in kw_args and kw_args['name'] is not None:
            raise ValueError(f'Name doubly specified: '
                             f'{name} and {kw_args["name"]}.')
        kw_args['name'] = name

        return method(*self.args, **kw_args)

    @abstractmethod
    def instantiate(self, vs, name, init):  # pragma: no cover
        """Instantiate the variable.

        Args:
            vs (:class:`.vars.Vars`): Variable container to extract the variable
                from.
            name (str): Name of the variable.
            init (object): Initial value.
        """


class Unbounded(VariableType):
    """Type of an unbounded variable."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.get, name, init)


class Positive(VariableType):
    """Type of a positive variable."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.positive, name, init)


class Bounded(VariableType):
    """Type of a bounded variable."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.bounded, name, init)


class LowerTriangular(VariableType):
    """Type of a lower-triangular matrix."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.lower_triangular, name, init)


class PositiveDefinite(VariableType):
    """Type of a positive-definite matrix."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.positive_definite, name, init)


class Orthogonal(VariableType):
    """Type of an orthogonal matrix."""

    def instantiate(self, vs, name, init):
        return self._get_var(vs.orthogonal, name, init)


def parametrised(prefix=None):
    """Decorator to specify variables with types.

    Args:
        prefix (str, optional): Prefix to prepend to the name of generated
            variables. Defaults to no prefix.
    """

    prefix, f = _extract_prefix_and_f(prefix)

    def decorator(f_):
        @wraps(f_)
        def wrapped_f(*args, **kw_args):
            signature = inspect.signature(f_)

            # Convert all arguments to keyword arguments, and store them here.
            filled_kwargs = {}

            # Look for variable container.
            values = args + tuple(kw_args.values())
            num_containers = sum([isinstance(x, Provider) for x in values])
            if num_containers == 0:
                raise ValueError('No variable container found.')
            elif num_containers > 1:
                raise ValueError('Multiple variable containers found.')
            else:
                # There is exactly only variable container. Find it.
                vs = [x for x in values if isinstance(x, Provider)][0]

            # Walk through the arguments.
            for name, parameter in signature.parameters.items():
                annotation = parameter.annotation

                # Instantiate uninstantiated variable types.
                if (isinstance(annotation, type) and
                        issubclass(annotation, VariableType)):
                    annotation = annotation()

                if isinstance(annotation, VariableType):
                    # Parameter is a variable that needs to be extracted.

                    # Check for default value.
                    if parameter.default is not parameter.empty:
                        init = parameter.default
                    else:
                        init = None

                    # Store the instantiated variable.
                    filled_kwargs[name] = \
                        annotation.instantiate(vs, prefix + name, init)

                else:
                    # Parameter is a regular parameter. Find it.

                    if len(args) > 0:
                        # Positional arguments left. Must be next one.
                        filled_kwargs[name] = args[0]
                        args = args[1:]
                    else:
                        # No position arguments left. Extract from keywords.
                        try:
                            filled_kwargs[name] = kw_args[name]
                            del kw_args[name]
                        except KeyError:
                            # Also not found in keywords. Resort to default
                            # value. If a default value is not given,
                            # the user did not specify a positional argument.
                            if parameter.default is parameter.empty:
                                raise ValueError(f'Positional argument '
                                                 f'"{name}" not given.')
                            else:
                                filled_kwargs[name] = parameter.default

            # Ensure that everything is parsed.
            if len(args) > 0:
                raise ValueError(f'{len(args)} positional argument(s) not '
                                 f'parsed.')
            if len(kw_args) > 0:
                raise ValueError(f'{len(kw_args)} keyword argument(s) not '
                                 f'parsed: {", ".join(kw_args.keys())}.')

            return f_(**filled_kwargs)

        return wrapped_f

    # Return the decorated function directly if a function was given.
    return decorator if f is None else decorator(f)



import logging

import tensorflow as tf

from ..minimise import make_l_bfgs_b, make_adam, exception

__all__ = ['minimise_l_bfgs_b', 'minimise_adam']

log = logging.getLogger(__name__)


def _wrap_f(vs, names, f):
    # Keep track of function evaluations.
    f_evals = []

    def f_wrapped(x):
        # Update to current point.
        vs.set_vector(x, *names)

        # Compute objective function value and gradient.
        try:
            with tf.GradientTape() as t:
                t.watch(vs.get_vars(*names))
                obj_value = f(vs)
                grads = t.gradient(obj_value, vs.get_vars(*names),
                                   unconnected_gradients='zero')
        except Exception as e:
            return exception(x, e)

        # Construct gradient.
        grad = vs.vector_packer.pack(*grads)

        f_evals.append(obj_value)
        return obj_value.numpy(), grad.numpy()

    return f_evals, f_wrapped


minimise_l_bfgs_b = make_l_bfgs_b(_wrap_f)
minimise_adam = make_adam(_wrap_f)



import numpy as np
import pytest
import torch
from varz import Vars
import lab as B

# noinspection PyUnresolvedReferences
from .util import (
    allclose,
    approx,
    KV,
    dtype,
    vs,
    vs_source,
    assert_lower_triangular,
    assert_positive_definite,
    assert_orthogonal
)


def test_get_vars():
    vs = Vars(np.int)

    # This test also tests that `Vars.get_vars` always returns the collection
    # of variables in the right order. This is important for optimisation.

    # Initialise some variables.
    vs.get(1, name='a')
    vs.get(2, name='1/b')
    vs.get(3, name='2/c')
    vs.unbounded(4, name='2/d')  # Test alias.

    # Test getting all.
    assert vs.get_vars() == [1, 2, 3, 4]
    assert vs.get_vars(indices=True) == [0, 1, 2, 3]

    # Test that names must exist.
    with pytest.raises(ValueError):
        vs.get_vars('e')

    # Test some queries.
    assert vs.get_vars('a') == [1]
    assert vs.get_vars('a', '*/b') == [1, 2]
    assert vs.get_vars('*/c', 'a') == [1, 3]
    assert vs.get_vars('*/c', '*/b', 'a') == [1, 2, 3]

    assert vs.get_vars('a', indices=True) == [0]
    assert vs.get_vars('a', '*/b', indices=True) == [0, 1]
    assert vs.get_vars('*/c', 'a', indices=True) == [0, 2]
    assert vs.get_vars('*/c', '*/b', 'a', indices=True) == [0, 1, 2]

    # Test some more queries.
    assert vs.get_vars('1/*') == [2]
    assert vs.get_vars('2/*') == [3, 4]
    assert vs.get_vars('1/*', '2/*') == [2, 3, 4]

    assert vs.get_vars('1/*', indices=True) == [1]
    assert vs.get_vars('2/*', indices=True) == [2, 3]
    assert vs.get_vars('1/*', '2/*', indices=True) == [1, 2, 3]

    # Test even more queries.
    assert vs.get_vars('*/b', '1/*') == [2]
    assert vs.get_vars('a', '2/*') == [1, 3, 4]
    assert vs.get_vars('a', '2/d', '2/*') == [1, 3, 4]
    assert vs.get_vars('2/d', '2/c', 'a', '1/*') == [1, 2, 3, 4]
    assert vs.get_vars('1/*') == [2]
    assert vs.get_vars('2/*') == [3, 4]
    assert vs.get_vars('1/*', '2/*') == [2, 3, 4]

    assert vs.get_vars('*/b', '1/*', indices=True) == [1]
    assert vs.get_vars('a', '2/*', indices=True) == [0, 2, 3]
    assert vs.get_vars('a', '2/d', '2/*', indices=True) == [0, 2, 3]
    assert vs.get_vars('2/d', '2/c', 'a', '1/*', indices=True) == [0, 1, 2, 3]
    assert vs.get_vars('1/*', indices=True) == [1]
    assert vs.get_vars('2/*', indices=True) == [2, 3]
    assert vs.get_vars('1/*', '2/*', indices=True) == [1, 2, 3]


def test_get_vars_cache_clearing(vs):
    vs.get(name='var_a')
    assert vs.get_vars('var_*', indices=True) == [0]
    vs.get(name='var_b')
    assert vs.get_vars('var_*', indices=True) == [0, 1]


def test_unbounded(vs_source):
    for _ in range(10):
        vs_source.get()
        vs_source.unbounded()


def test_unbounded_init(vs):
    vs.get(1, name='x')
    allclose(vs['x'], 1)


def test_unbounded_assignment(vs):
    vs.get(1, name='x')
    vs.assign('x', 2)
    allclose(vs['x'], 2)


def test_positive(vs_source):
    for _ in range(10):
        assert vs_source.pos() >= 0
        assert vs_source.positive() >= 0


def test_positive_init(vs):
    vs.pos(1, name='x')
    allclose(vs['x'], 1)


def test_positive_assignment(vs):
    vs.pos(1, name='x')
    vs.assign('x', 2)
    allclose(vs['x'], 2)


def test_bounded(vs_source):
    for _ in range(10):
        assert 10 <= vs_source.bnd(lower=10, upper=11) <= 11
        assert 10 <= vs_source.bounded(lower=10, upper=11) <= 11


def test_bounded_init(vs):
    vs.bnd(2, name='x', lower=1, upper=4)
    allclose(vs['x'], 2)


def test_bounded_assignment(vs):
    vs.bnd(2, name='x', lower=1, upper=4)
    vs.assign('x', 3)
    allclose(vs['x'], 3)


def test_lower_triangular(vs_source):
    for _ in range(10):
        assert B.shape(vs_source.tril(shape=(5, 5))) == (5, 5)
        assert_lower_triangular(vs_source.tril(shape=(5, 5)))
        assert_lower_triangular(vs_source.lower_triangular(shape=(5, 5)))


def test_lower_triangular_init(vs):
    x = vs.tril(shape=(5, 5))

    vs.tril(x, name='x')
    allclose(vs['x'], x)

    with pytest.raises(ValueError):
        vs.tril()
    with pytest.raises(ValueError):
        vs.tril(shape=5)
    with pytest.raises(ValueError):
        vs.tril(shape=(5, 6))


def test_lower_triangular_assignment(vs):
    x = vs.tril(shape=(5, 5))

    vs.tril(shape=(5, 5), name='x')
    vs.assign('x', x)
    allclose(vs['x'], x)


def test_positive_definite(vs_source):
    for _ in range(10):
        assert B.shape(vs_source.pd(shape=(5, 5))) == (5, 5)
        assert_positive_definite(vs_source.pd(shape=(5, 5)))
        assert_positive_definite(vs_source.positive_definite(shape=(5, 5)))


def test_positive_definite_init(vs):
    x = vs.pd(shape=(5, 5))

    vs.pd(x, name='x')
    allclose(vs['x'], x)

    with pytest.raises(ValueError):
        vs.pd()
    with pytest.raises(ValueError):
        vs.pd(shape=5)
    with pytest.raises(ValueError):
        vs.pd(shape=(5, 6))


def test_positive_definite_assignment(vs):
    x = vs.pd(shape=(5, 5))

    vs.pd(shape=(5, 5), name='x')
    vs.assign('x', x)
    allclose(vs['x'], x)


def test_orthogonal(vs_source):
    for i in range(10):
        assert B.shape(vs_source.orth(shape=(5, 5))) == (5, 5)
        assert_orthogonal(vs_source.orth(shape=(5, 5)))
        assert_orthogonal(vs_source.orthogonal(shape=(5, 5)))


def test_orthogonal_init(vs):
    x = vs.orth(shape=(5, 5))

    vs.orth(x, name='x')
    allclose(vs['x'], x)

    with pytest.raises(ValueError):
        vs.orth()
    with pytest.raises(ValueError):
        vs.orth(shape=5)
    with pytest.raises(ValueError):
        vs.orth(shape=(5, 6))


def test_orthogonal_assignment(vs):
    x = vs.orth(shape=(5, 5))

    vs.orth(shape=(5, 5), name='x')
    vs.assign('x', x)
    allclose(vs['x'], x)


def test_get_set_vector(dtype):
    vs = Vars(dtype=dtype)

    # Test stacking a matrix and a vector.
    vs.get(shape=(2,), name='a', init=np.array([1, 2]))
    vs.get(shape=(2, 2), name='b', init=np.array([[3, 4], [5, 6]]))
    allclose(vs.get_vector('a', 'b'), [1, 2, 3, 4, 5, 6])

    # Test setting elements.
    vs.set_vector(np.array([6, 5, 4, 3, 2, 1]), 'a', 'b')
    allclose(vs['a'], np.array([6, 5]))
    allclose(vs['b'], np.array([[4, 3], [2, 1]]))

    # Test setting elements in a differentiable way. This should allow for
    # any values.
    vs.set_vector(np.array(['1', '2', '3', '4', '5', '6']), 'a', 'b',
                  differentiable=True)
    assert np.all(vs['a'] == ['1', '2'])
    assert np.all(vs['b'] == np.array([['3', '4'], ['5', '6']]))


def test_differentiable_assignment(vs):
    # Differentiably assigning new values should allow for anything.
    vs.get(name='x')
    vs.assign('x', 'value', differentiable=True)
    assert vs['x'] == 'value'


def test_copy_torch():
    vs = Vars(torch.float64)

    # Create a variable.
    vs.pos(1, name='a')

    # Initialise vector packer.
    vs.get_vector()

    # Make a normal and detached copy.
    vs_copy = vs.copy()
    vs_detached = vs.copy(detach=True)

    # Require gradients for both.
    vs.requires_grad(True)
    vs_detached.requires_grad(True)

    # Do a backward pass.
    (vs_detached['a'] ** 2).backward()

    # Check that values are equal, but gradients only computed for one.
    assert vs['a'] == 1
    assert vs.get_vars('a')[0].grad is None
    assert vs_detached['a'] == 1
    assert vs_detached.get_vars('a')[0].grad == 2

    # Check that copied fields are, in fact, copies, and that the vector packer
    # is also copied.
    del vs.transforms[:]
    del vs.inverse_transforms[:]
    del vs.vars[:]
    vs.name_to_index.clear()

    for vs_copied in [vs_copy, vs_detached]:
        assert len(vs_copied.transforms) > 0
        assert len(vs_copied.inverse_transforms) > 0
        assert len(vs_copied.vars) > 0
        assert len(vs_copied.name_to_index) > 0
        assert vs_copied.vector_packer != None


def test_requires_grad_detach_vars_torch():
    vs = Vars(torch.float64)
    vs.pos(1, name='a')

    # Test that gradients need to first be required.
    with pytest.raises(RuntimeError):
        (2 * vs['a']).backward()

    # Test that gradients can be required and are then computed.
    vs.requires_grad(True)
    (2 * vs['a']).backward()
    assert vs.vars[0].grad is not None

    # Test that variables can be detached.
    vs.pos(1, name='b')
    result = 2 * vs['b']
    vs.detach()
    with pytest.raises(RuntimeError):
        result.backward()


def test_source():
    vs = Vars(np.float32, source=np.ones(10))

    assert vs.get() == 1.
    approx(vs.pos(shape=(5,)), np.exp(np.ones(5)))
    approx(vs.pos(), np.exp(1.))
    with pytest.raises(ValueError):
        vs.pos(shape=(5,))

    # Test that the source variables are casted to the right data type.

    vs = Vars(np.float32, source=np.array([1]))
    assert vs.get().dtype == np.float32

    vs = Vars(np.float64, source=np.array([1]))
    assert vs.get().dtype == np.float64


def test_names(vs):
    assert vs.names == []
    vs.get()
    assert vs.names == []
    vs.pos(name='a')
    assert vs.names == ['a']
    vs.bnd(name='b')
    assert vs.names == ['a', 'b']
    vs.get(name='c')
    assert vs.names == ['a', 'b', 'c']


def test_print(vs):
    with KV() as mock:
        vs.print()
        assert mock.keys == []
        assert mock.values == []

        vs.get()
        vs.print()
        assert mock.keys == []
        assert mock.values == []

        vs.print()
        vs.get(1, name='a')
        vs.print()
        assert mock.keys == ['a']
        assert mock.values == [1]

        vs.get(2, name='b')
        vs.print()
        assert mock.keys == ['a', 'a', 'b']
        assert mock.values == [1, 1, 2]

        vs.get(3, name='c')
        vs.print()
        assert mock.keys == ['a', 'a', 'b', 'a', 'b', 'c']
        assert mock.values == [1, 1, 2, 1, 2, 3]



import pytest

from varz import (
    sequential,
    Unbounded,
    Positive,
    Bounded,
    LowerTriangular,
    PositiveDefinite,
    Orthogonal,
    parametrised
)
from varz.spec import _extract_prefix_and_f
# noinspection PyUnresolvedReferences
from .util import (
    vs,
    assert_lower_triangular,
    assert_positive_definite,
    assert_orthogonal
)


def test_extract_prefix_and_f():
    def f():
        pass

    assert _extract_prefix_and_f(None) == ('', None)
    assert _extract_prefix_and_f(f) == ('', f)
    assert _extract_prefix_and_f('test') == ('test', None)


@pytest.mark.parametrize('decorator, names',
                         [(sequential, ['z', '0', '1', '2']),
                          (sequential(), ['z', '0', '1', '2']),
                          (sequential('x'), ['z', 'x0', 'x1', 'x2'])])
def test_sequential(vs, decorator, names):
    vs.get(0, name='z')

    @decorator
    def f(x, vs_, y):
        assert x == 1
        assert y == 2
        return vs_['z'], vs_.get(), vs_.pos(), vs_.bnd(lower=10, upper=11)

    # Test that the same variables are retrieved.
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names

    # Test correctness off variables.
    assert f(1, vs, 2)[0] == 0
    assert f(1, vs, 2)[2] >= 0
    assert 10 <= f(1, vs, 2)[3] <= 11


def test_sequential_unbounded(vs):
    @sequential
    def f(vs_):
        return vs_.get(1)

    assert f(vs) == 1


def test_sequential_positive(vs):
    @sequential
    def f(vs_):
        return vs_.pos()

    assert f(vs) > 0


def test_sequential_bounded(vs):
    @sequential
    def f(vs_):
        return vs_.bnd(lower=10, upper=11)

    assert 10 <= f(vs) <= 11


def test_sequential_lower_triangular(vs):
    @sequential
    def f(vs_):
        return vs_.tril(shape=(5, 5))

    assert_lower_triangular(f(vs))


def test_sequential_positive_definite(vs):
    @sequential
    def f(vs_):
        return vs_.pd(shape=(5, 5))

    assert_positive_definite(f(vs))


def test_sequential_orthogonal(vs):
    @sequential
    def f(vs_):
        return vs_.orth(shape=(5, 5))

    assert_orthogonal(f(vs))


@pytest.mark.parametrize('decorator, types, names',
                         [(parametrised,
                           [Unbounded, Positive, Bounded(lower=10, upper=11)],
                           ['w', 'x', 'y', 'z']),
                          (parametrised(),
                           [Unbounded, Positive, Bounded(lower=10, upper=11)],
                           ['w', 'x', 'y', 'z']),
                          (parametrised('var_'),
                           [Unbounded, Positive, Bounded(lower=10, upper=11)],
                           ['w', 'var_x', 'var_y', 'var_z'])])
def test_parametrised(vs, decorator, types, names):
    vs.get(0, name='w')

    @decorator
    def f(a, x: types[0], vs_, y: types[1], b, z: types[2], c=None):
        assert a == 1
        assert b == 2
        assert c is None
        return vs_['w'], x, y, z

    # Test that the same variables are retrieved
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names
    assert f(1, vs, 2) == f(1, vs, 2)
    assert vs.names == names

    # Test correctness off variables.
    assert f(1, vs, 2)[0] == 0
    assert f(1, vs, 2)[2] >= 0
    assert 10 <= f(1, vs, 2)[3] <= 11


def test_parametrised_arguments(vs):
    @parametrised
    def f(vs_, x: Unbounded, a, y: Unbounded, b=2, z: Unbounded = 10, c=3):
        return a, b, c

    with pytest.raises(ValueError):
        f(vs)

    assert f(vs, 1) == (1, 2, 3)
    assert f(vs, a=1) == (1, 2, 3)

    assert f(vs, 1, 4) == (1, 4, 3)
    assert f(vs, 1, b=4) == (1, 4, 3)
    assert f(vs, a=1, b=4) == (1, 4, 3)

    assert f(vs, 1, 4, 5) == (1, 4, 5)
    assert f(vs, 1, 4, c=5) == (1, 4, 5)
    assert f(vs, a=1, b=4, c=5) == (1, 4, 5)

    # Test unspecified positional argument.
    with pytest.raises(ValueError):
        f(vs, 1, 2, 3, 4)

    # Test unspecified keyword argument.
    with pytest.raises(ValueError):
        f(vs, 1, 2, d=3)


def test_parametrised_variable_container(vs):
    @parametrised
    def f(x, y, z: Unbounded):
        pass

    # Test that a variable container must be present.
    with pytest.raises(ValueError):
        f(1, 2)

    # Test that only one variable container can be present.
    with pytest.raises(ValueError):
        f(vs, vs)


def test_parametrised_double_initial_value(vs):
    @parametrised
    def f(vs_, x: Unbounded(init=5) = 5):
        pass

    with pytest.raises(ValueError):
        f(vs)


def test_parametrised_double_name(vs):
    @parametrised
    def f(vs_, x: Unbounded(name='x')):
        pass

    with pytest.raises(ValueError):
        f(vs)


def test_parametrised_unbounded(vs):
    @parametrised
    def f(vs_, x: Unbounded = 1):
        return x

    assert f(vs) == 1


def test_parametrised_positive(vs):
    @parametrised
    def f(vs_, x: Positive):
        return x

    assert f(vs) > 0


def test_parametrised_bounded(vs):
    @parametrised
    def f(vs_, x: Bounded(lower=10, upper=11)):
        return x

    assert 10 <= f(vs) <= 11


def test_parametrised_lower_triangular(vs):
    @parametrised
    def f(vs_, x: LowerTriangular(shape=(5, 5))):
        return x

    assert_lower_triangular(f(vs))


def test_parametrised_positive_definite(vs):
    @parametrised
    def f(vs_, x: PositiveDefinite(shape=(5, 5))):
        return x

    assert_positive_definite(f(vs))


def test_parametrised_orthogonal(vs):
    @parametrised
    def f(vs_, x: Orthogonal(shape=(5, 5))):
        return x

    assert_orthogonal(f(vs))



import logging

import torch

from ..minimise import make_l_bfgs_b, make_adam, exception

__all__ = ['minimise_l_bfgs_b', 'minimise_adam']

log = logging.getLogger(__name__)


def _wrap_f(vs, names, f):
    # Differentiable assignments will overwrite the variables, so make a copy
    # with detached variables.
    vs_copy = vs.copy(detach=True)

    # Keep track of function evaluations.
    f_evals = []

    def f_wrapped(x):
        x_torch = torch.tensor(x, requires_grad=True)

        # Compute objective function value.
        try:
            vs_copy.set_vector(x_torch, *names, differentiable=True)
            obj_value = f(vs_copy)
            obj_value.backward()
            obj_value = obj_value.detach_().numpy()
        except Exception as e:
            return exception(x, e)

        # Extract gradient.
        grad = x_torch.grad.detach_().numpy()

        f_evals.append(obj_value)
        return obj_value, grad

    return f_evals, f_wrapped


minimise_l_bfgs_b = make_l_bfgs_b(_wrap_f)
minimise_adam = make_adam(_wrap_f)



from itertools import product

import lab as B
import numpy as np
from varz import Vars, Initialiser, Packer

from .util import allclose, vs


def test_initialiser(vs):
    init = Initialiser()

    # Initialise two variables.
    vs.get(1., name='a')
    vs.pos(2., name='b')

    # Define some initialisations.
    init.assign('a', [-3., 4.])
    init.assign('b', [5., 6.])
    inits = init.generate(vs)

    # Test the initialisations.
    for initialiser, values in zip(inits, product([-3., 4.], [5., 6.])):
        initialiser()
        assert vs['a'] == values[0]
        allclose(vs['b'], values[1])


def test_packer():
    a, b, c = B.randn(5, 10), B.randn(20), B.randn(5, 1, 15)

    for packer, args in zip([Packer(a, b, c), Packer([a, b, c])],
                            [(a, b, c), ((a, b, c),)]):
        # Test packing.
        packed = packer.pack(*args)
        assert B.rank(packed) == 1

        # Test unpacking.
        a_, b_, c_ = packer.unpack(packed)
        allclose(a, a_)
        allclose(b, b_)
        allclose(c, c_)



import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from functools import reduce
from operator import mul

import lab as B
import numpy as np
import wbml.out
from plum import Dispatcher, Self, Referentiable

from .util import Packer, match, lazy_tf as tf, lazy_torch as torch

__all__ = ['Provider', 'Vars']

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


@_dispatch(B.NPNumeric, B.Numeric)
def _assign(x, value):
    np.copyto(x, value)
    return x


@_dispatch(B.TFNumeric, B.Numeric)
def _assign(x, value):
    return x.assign(value)


@_dispatch(B.TorchNumeric, B.Numeric)
def _assign(x, value):
    if not isinstance(value, B.TorchNumeric):
        value = torch.tensor(value, dtype=x.dtype)
    x.data.copy_(value)
    return x


class Provider(metaclass=Referentiable(ABCMeta)):
    @abstractmethod
    def unbounded(self,
                  init=None,
                  shape=(),
                  dtype=None,
                  name=None):  # pragma: no cover
        """Get an unbounded variable.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (tuple[int], optional): Shape of the variable. Defaults to
                scalar.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (str, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def get(self, *args, **kw_args):
        """Alias for :meth:`.vars.Provider.unbounded`."""
        return self.unbounded(*args, **kw_args)

    @abstractmethod
    def positive(self,
                 init=None,
                 shape=(),
                 dtype=None,
                 name=None):  # pragma: no cover
        """Get a positive variable.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (tuple[int], optional): Shape of the variable. Defaults to
                scalar.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (str, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def pos(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.positive`."""
        return self.positive(*args, **kw_args)

    @abstractmethod
    def bounded(self,
                init=None,
                lower=1e-4,
                upper=1e4,
                shape=(),
                dtype=None,
                name=None):  # pragma: no cover
        """Get a bounded variable.

        Args:
            init (tensor, optional): Initialisation of the variable.
            lower (tensor, optional): Lower bound. Defaults to `1e-4`.
            upper (tensor, optional): Upper bound. Defaults to `1e4`.
            shape (tuple[int], optional): Shape of the variable. Defaults to
                scalar.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def bnd(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.bounded`."""
        return self.bounded(*args, **kw_args)

    @abstractmethod
    def lower_triangular(self,
                         init=None,
                         shape=None,
                         dtype=None,
                         name=None):  # pragma: no cover
        """Get a lower-triangular matrix.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (int, optional): Number of rows and columns of the matrix.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def tril(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.lower_triangular`."""
        return self.lower_triangular(*args, **kw_args)

    @abstractmethod
    def positive_definite(self,
                          init=None,
                          shape=None,
                          dtype=None,
                          name=None):  # pragma: no cover
        """Get a positive-definite matrix.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (int, optional): Number of rows and columns of the matrix.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def pd(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.positive_definite`."""
        return self.positive_definite(*args, **kw_args)

    @abstractmethod
    def orthogonal(self,
                   init=None,
                   shape=None,
                   dtype=None,
                   name=None):  # pragma: no cover
        """Get an orthogonal matrix.

        Args:
            init (tensor, optional): Initialisation of the variable.
            shape (int, optional): Number of rows and columns of the matrix.
            dtype (data type, optional): Data type of the variable. Defaults to
                that of the storage.
            name (hashable, optional): Name of the variable.

        Returns:
            tensor: Variable.
        """

    def orth(self, *args, **kw_args):
        """Alias for :meth:`.vars.Vars.orthogonal`."""
        return self.orthogonal(*args, **kw_args)

    @abstractmethod
    def __getitem__(self, name):  # pragma: no cover
        """Get a variable by name.

        Args:
            name (hashable): Name of variable.

        Returns:
            tensor: Variable.
        """


@_dispatch(object)
def _check_matrix_shape(shape):
    raise ValueError(f'Object {shape} is not a shape.')


@_dispatch(tuple)
def _check_matrix_shape(shape):
    if len(shape) != 2 or shape[0] != shape[1]:
        raise ValueError(f'Shape {shape} must be the shape of a square matrix.')


def _check_init_shape(init, shape):
    if init is None and shape is None:
        raise ValueError(f'The shape must be given to automatically initialise '
                         f'a matrix variable.')
    if shape is None:
        shape = B.shape(init)
    return init, shape


class Vars(Provider):
    """Variable storage.

    Args:
        dtype (data type): Data type of the variables.
        source (tensor, optional): Tensor to source variables from. Defaults to
            not being used.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, dtype, source=None):
        self.dtype = dtype

        # Source:
        self.source = source
        self.source_index = 0

        # Storage:
        self.vars = []
        self.transforms = []
        self.inverse_transforms = []

        # Lookup:
        self.name_to_index = OrderedDict()
        self._get_vars_cache = {}

        # Packing:
        self.vector_packer = None

    def _get_var(self,
                 transform,
                 inverse_transform,
                 init,
                 generate_init,
                 shape,
                 shape_latent,
                 dtype,
                 name):
        # If the name already exists, return that variable.
        try:
            return self[name]
        except KeyError:
            pass

        # A new variable will be added. Clear lookup cache.
        self._get_vars_cache.clear()

        # Resolve data type.
        dtype = self.dtype if dtype is None else dtype

        # If no source is provided, get the latent from from the provided
        # initialiser.
        if self.source is None:
            # Resolve initialisation and inverse transform.
            if init is None:
                init = generate_init(shape=shape, dtype=dtype)
            else:
                init = B.cast(dtype, init)

            # Construct optimisable variable.
            latent = inverse_transform(init)
            if isinstance(self.dtype, B.TFDType):
                latent = tf.Variable(latent)
            elif isinstance(self.dtype, B.TorchDType):
                pass  # All is good in this case.
            else:
                # Must be a NumPy data type.
                assert isinstance(self.dtype, B.NPDType)
                latent = np.array(latent)
        else:
            # Get the latent variable from the source.
            length = reduce(mul, shape_latent, 1)
            latent_flat = \
                self.source[self.source_index:self.source_index + length]
            self.source_index += length

            # Cast to the right data type.
            latent = B.cast(dtype, B.reshape(latent_flat, *shape_latent))

        # Store transforms.
        self.vars.append(latent)
        self.transforms.append(transform)
        self.inverse_transforms.append(inverse_transform)

        # Get index of the variable.
        index = len(self.vars) - 1

        # Store name if given.
        if name is not None:
            self.name_to_index[name] = index

        # Generate the variable and return.
        return transform(latent)

    def unbounded(self, init=None, shape=(), dtype=None, name=None):
        def generate_init(shape, dtype):
            return B.randn(dtype, *shape)

        return self._get_var(transform=lambda x: x,
                             inverse_transform=lambda x: x,
                             init=init,
                             generate_init=generate_init,
                             shape=shape,
                             shape_latent=shape,
                             dtype=dtype,
                             name=name)

    def positive(self, init=None, shape=(), dtype=None, name=None):
        def generate_init(shape, dtype):
            return B.rand(dtype, *shape)

        return self._get_var(transform=lambda x: B.exp(x),
                             inverse_transform=lambda x: B.log(x),
                             init=init,
                             generate_init=generate_init,
                             shape=shape,
                             shape_latent=shape,
                             dtype=dtype,
                             name=name)

    def bounded(self,
                init=None,
                lower=1e-4,
                upper=1e4,
                shape=(),
                dtype=None,
                name=None):
        def transform(x):
            return lower + (upper - lower) / (1 + B.exp(x))

        def inverse_transform(x):
            return B.log(upper - x) - B.log(x - lower)

        def generate_init(shape, dtype):
            return lower + B.rand(dtype, *shape) * (upper - lower)

        return self._get_var(transform=transform,
                             inverse_transform=inverse_transform,
                             init=init,
                             generate_init=generate_init,
                             shape=shape,
                             shape_latent=shape,
                             dtype=dtype,
                             name=name)

    def lower_triangular(self,
                         init=None,
                         shape=None,
                         dtype=None,
                         name=None):
        init, shape = _check_init_shape(init, shape)
        _check_matrix_shape(shape)

        def transform(x):
            return B.vec_to_tril(x)

        def inverse_transform(x):
            return B.tril_to_vec(x)

        def generate_init(shape, dtype):
            mat = B.randn(dtype, *shape)
            return transform(B.tril_to_vec(mat))

        shape_latent = (int(shape[0] * (shape[0] + 1) / 2),)
        return self._get_var(transform=transform,
                             inverse_transform=inverse_transform,
                             init=init,
                             generate_init=generate_init,
                             shape=shape,
                             shape_latent=shape_latent,
                             dtype=dtype,
                             name=name)

    def positive_definite(self,
                          init=None,
                          shape=None,
                          dtype=None,
                          name=None):
        init, shape = _check_init_shape(init, shape)
        _check_matrix_shape(shape)

        def transform(x):
            chol = B.vec_to_tril(x)
            return B.matmul(chol, chol, tr_b=True)

        def inverse_transform(x):
            chol = B.cholesky(B.reg(x))
            return B.tril_to_vec(chol)

        def generate_init(shape, dtype):
            mat = B.randn(dtype, *shape)
            return transform(B.tril_to_vec(mat))

        shape_latent = (int(shape[0] * (shape[0] + 1) / 2),)
        return self._get_var(transform=transform,
                             inverse_transform=inverse_transform,
                             init=init,
                             generate_init=generate_init,
                             shape=shape,
                             shape_latent=shape_latent,
                             dtype=dtype,
                             name=name)

    def orthogonal(self,
                   init=None,
                   shape=None,
                   dtype=None,
                   name=None):
        init, shape = _check_init_shape(init, shape)
        _check_matrix_shape(shape)

        def transform(x):
            tril = B.vec_to_tril(x)
            skew = tril - B.transpose(tril)
            eye = B.eye(skew)
            return B.solve(eye + skew, eye - skew)

        def inverse_transform(x):
            eye = B.eye(x)
            return B.tril_to_vec(B.solve(eye + x, eye - x))

        def generate_init(shape, dtype):
            mat = B.randn(dtype, *shape)
            return transform(B.tril_to_vec(mat))

        shape_latent = (int(shape[0] * (shape[0] + 1) / 2),)
        return self._get_var(transform=transform,
                             inverse_transform=inverse_transform,
                             init=init,
                             generate_init=generate_init,
                             shape=shape,
                             shape_latent=shape_latent,
                             dtype=dtype,
                             name=name)

    def __getitem__(self, name):
        index = self.name_to_index[name]
        return self.transforms[index](self.vars[index])

    def assign(self, name, value, differentiable=False):
        """Assign a value to a variable.

        Args:
            name (hashable): Name of variable to assign value to.
            value (tensor): Value to assign.
            differentiable (bool, optional): Do a differentiable assignment.

        Returns:
            tensor: Assignment result.
        """
        index = self.name_to_index[name]
        if differentiable:
            # Do a differentiable assignment.
            self.vars[index] = value
            return value
        else:
            # Overwrite data.
            return _assign(self.vars[index],
                           self.inverse_transforms[index](value))

    def copy(self, detach=False):
        """Create a copy of the variable manager that shares the variables.

        Args:
            detach (bool, optional): Detach the variables in PyTorch. Defaults
                to `False`.

        Returns:
            :class:`.vars.Vars`: Copy.
        """
        vs = Vars(dtype=self.dtype)
        vs.transforms = list(self.transforms)
        vs.inverse_transforms = list(self.inverse_transforms)
        vs.name_to_index = OrderedDict(self.name_to_index)
        vs.vector_packer = self.vector_packer
        if detach:
            for var in self.vars:
                vs.vars.append(var.detach())
        else:
            vs.vars = list(self.vars)
        return vs

    def detach(self):
        """Detach all variables held in PyTorch."""
        self.vars = [v.detach() for v in self.vars]

    def requires_grad(self, value, *names):
        """Set which variables require a gradient in PyTorch.

        Args:
            value (bool): Require a gradient.
            *names (hashable): Specify variables by name.
        """
        for var in self.get_vars(*names):
            var.requires_grad_(value)

    def get_vars(self, *names, **kw_args):
        """Get latent variables.

        If no arguments are supplied, then all latent variables are retrieved.
        Furthermore, the same collection of variables is guaranteed to be
        returned in the same order.

        Args:
            *names (hashable): Get variables by name.
            indices (bool, optional): Get the indices of the variables instead.
                Defaults to `False`.

        Returns:
            list: Matched latent variables or their indices, depending on the
                value of `indices`.
        """
        # If nothing is specified, return all latent variables.
        if len(names) == 0:
            if kw_args.get('indices', False):
                return list(range(len(self.vars)))
            else:
                return self.vars

        # Attempt to use cache.
        cache_key = (names, kw_args.get('indices', False))
        try:
            return self._get_vars_cache[cache_key]
        except KeyError:
            pass

        # Collect indices of matches.
        indices = set()
        for name in names:
            a_match = False
            for k, v in self.name_to_index.items():
                if match(name, k):
                    indices |= {v}
                    a_match = True

            # Check that there was a match.
            if not a_match:
                raise ValueError('No variable matching "{}".'.format(name))

        # Return indices if asked for. Otherwise, return variables.
        if kw_args.get('indices', False):
            res = sorted(indices)
        else:
            res = [self.vars[i] for i in sorted(indices)]

        # Store in cache before returning.
        self._get_vars_cache[cache_key] = res
        return res

    def get_vector(self, *names):
        """Get all the latent variables stacked in a vector.

        If no arguments are supplied, then all latent variables are retrieved.

        Args:
            *names (hashable): Get variables by name.

        Returns:
            tensor: Vector consisting of all latent values
        """
        vars = self.get_vars(*names)
        self.vector_packer = Packer(*vars)
        return self.vector_packer.pack(*vars)

    def set_vector(self, values, *names, **kw_args):
        """Set all the latent variables by values from a vector.

        If no arguments are supplied, then all latent variables are retrieved.

        Args:
            values (tensor): Vector to set the variables to.
            *names (hashable): Set variables by name.
            differentiable (bool, optional): Differentiable assignment. Defaults
                to `False`.

        Returns:
            list: Assignment results.
        """
        values = self.vector_packer.unpack(values)

        if kw_args.get('differentiable', False):
            # Do a differentiable assignment.
            for index, value in zip(self.get_vars(*names, indices=True),
                                    values):
                self.vars[index] = value
            return values
        else:
            # Overwrite data.
            assignments = []
            for var, value in zip(self.get_vars(*names), values):
                assignments.append(_assign(var, value))
            return assignments

    @property
    def names(self):
        """All available names."""
        return list(self.name_to_index.keys())

    def print(self):
        """Print all variables."""
        for name in self.names:
            wbml.out.kv(name, self[name])



import lab as B
import numpy as np
import pytest
import tensorflow as tf
import torch
import wbml.out
from numpy.testing import assert_allclose, assert_array_almost_equal
from plum import Dispatcher
from wbml import out as out
from varz import Vars

__all__ = ['Value',
           'allclose',
           'approx',

           # Numerical checks:
           'assert_lower_triangular',
           'assert_positive_definite',
           'assert_orthogonal',

           # Fixtures:
           'dtype',
           'vs',
           'vs_source',

           # Mocks:
           'KV',
           'OutStream']

_dispatch = Dispatcher()


class Value:
    def __init__(self, val):
        self.val = val


@_dispatch({B.Number, B.NP, list})
def _to_numpy(x):
    return x


@_dispatch({B.Torch, B.TF})
def _to_numpy(x):
    return x.numpy()


def approx(x, y, digits=7):
    x = _to_numpy(x)
    y = _to_numpy(y)
    assert_array_almost_equal(x, y, decimal=digits)


def allclose(x, y):
    x = _to_numpy(x)
    y = _to_numpy(y)
    assert_allclose(x, y)


# Numerical checks:

def assert_lower_triangular(x):
    """Assert that a matrix is lower triangular."""
    # Check that matrix is square.
    assert B.shape(x)[0] == B.shape(x)[1]

    # Check that upper part is all zeros.
    upper = x[np.triu_indices(B.shape(x)[0], k=1)]
    allclose(upper, B.zeros(upper))


def assert_positive_definite(x):
    """Assert that a matrix is positive definite."""
    # Check that Cholesky decomposition succeeds.
    B.cholesky(x)


def assert_orthogonal(x):
    """Assert that a matrix is orthogonal."""
    # Check that matrix is square.
    assert B.shape(x)[0] == B.shape(x)[1]

    # Check that its transpose is its inverse.
    approx(B.matmul(x, x, tr_a=True), B.eye(x))


# Fixtures:

@pytest.fixture(params=[np.float64, torch.float64, tf.float64])
def dtype(request):
    return request.param


@pytest.fixture()
def vs():
    return Vars(np.float64)


@pytest.fixture(params=[False, True])
def vs_source(request):
    if request.param:
        source = B.randn(np.float64, 1000)
        return Vars(np.float64, source=source)
    else:
        return Vars(np.float64)


# Mocks:


class KV:
    """Mock `wbml.out.kv`."""

    def __init__(self):
        self.keys = []
        self.values = []
        self._kv = None

    def __call__(self, key, value):
        self.keys.append(key)
        self.values.append(value)

    def __enter__(self):
        self._kv = wbml.out.kv
        wbml.out.kv = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        wbml.out.kv = self._kv


class OutStream:
    """Mock the streams of `wbml.out`."""

    def __init__(self):
        self.output = ''

    def write(self, msg):
        self.output += msg

    def __enter__(self):
        self._orig_streams = list(out.streams)
        out.streams[:] = [self]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        out.streams[:] = self._orig_streams



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from setuptools import find_packages, setup

requirements = ['numpy>=1.16',
                'scipy>=1.3',

                'plum-dispatch',
                'backends>=0.3',
                'wbml']

setup(packages=find_packages(exclude=['docs']),
      python_requires='>=3.6',
      install_requires=requirements,
      include_package_data=True)
